/* handle display, keyboard, joystick, and audio in emerald mine */

/* uses allegro low level game routines */

extern char *global;
static char *local = "allegro";

#include "proto.h"

#define ALLEGRO_NO_COMPATIBILITY
#include <allegro.h>

#include <string.h>
#include <stdio.h>

/* ---------------------------------------------------------------------- */

static void parseGeometry(char *parsestring, int *x_return, int *y_return, unsigned int *width_return, unsigned int *height_return)
{
	int x = 0, y = 0;
	unsigned int width = 0, height = 0;
	char gotxy = 0, gotwh = 0;
	if(*parsestring == '=') parsestring++;
	if(*parsestring != '+' && *parsestring != '-') {
		char *endptr;
		width = ustrtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		if(*parsestring != 'x' && *parsestring != 'X') return;
		parsestring++;
		height = ustrtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		gotwh = 1;
	}
	if(*parsestring == '+' || *parsestring == '-') {
		char *endptr;
		parsestring++;
		x = ustrtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		if(*parsestring != '+' && *parsestring != '-') return;
		parsestring++;
		y = ustrtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		gotxy = 1;
	}
	if(*parsestring != 0) return;

	if(gotxy) {
		if(x_return) *x_return = x;
		if(y_return) *y_return = y;
	}
	if(gotwh) {
		if(width_return) *width_return = width;
		if(height_return) *height_return = height;
	}
}

/* ---------------------------------------------------------------------- */

/* initialize allegro, create window, create palette */

static int scale_x = 2; /* default to double size because it looks better on most displays */
static int scale_y = 2;

static int center_dx, center_dy;

static PALETTE privateColours;
static unsigned long paletteHash[128];

int display_open(void)
{
	int i;

	if(install_allegro(SYSTEM_AUTODETECT, 0, 0)) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to initialize system", allegro_error);
		return(1);
	}
	if(install_timer()) {
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to initialize timer", allegro_error);
		return(1);
	}
	if(install_keyboard()) {
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to initialize keyboard", allegro_error);
		return(1);
	}
	if(install_mouse() == -1) {
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to initialize mouse", allegro_error);
		return(1);
	}
	if(install_joystick(JOY_TYPE_AUTODETECT))
	if(install_joystick(JOY_TYPE_NONE))
	{
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to initialize joystick", allegro_error);
		return(1);
	}
	if(install_sound(DIGI_AUTODETECT, MIDI_NONE, 0))
	if(install_sound(DIGI_NONE, MIDI_NONE, 0))
	{
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to initialize audio", allegro_error);
		return(1);
	}

	for(i = 0; i < COLOUR_MAX; i++) {
		unsigned long rgbc = colours_xpm[i];
		privateColours[i + 1].r = ((rgbc >> 24) & 255) >> 2;
		privateColours[i + 1].g = ((rgbc >> 16) & 255) >> 2;
		privateColours[i + 1].b = ((rgbc >> 8) & 255) >> 2;
		paletteHash[rgbc & 127] = i + 1;
	}

	if(arg_geometry) {
		int dummy;
		unsigned int w = scale_x;
		unsigned int h = scale_y;
		parseGeometry(arg_geometry, &dummy, &dummy, &w, &h);
		scale_x = w;
		scale_y = h;
	}
	if(scale_x < 1 || scale_x > 9 || scale_y < 1 || scale_y > 9) {
		scale_x = 1;
		scale_y = 1;
	}

	set_color_depth(8);
	request_refresh_rate(60);

	if(set_gfx_mode((arg_fullscreen ? GFX_AUTODETECT_FULLSCREEN : GFX_AUTODETECT_WINDOWED), VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y, 0, 0))
#ifdef GFX_MODEX
	if(scale_x = 1, scale_y = 1, set_gfx_mode(GFX_MODEX, VIEW_WIDTH, VIEW_HEIGHT, 0, 0))
#endif
	if(scale_x = 1, scale_y = 1, set_gfx_mode(GFX_SAFE, 320, 200, 0, 0))
	{
		allegro_message("%s[%s] %s: %s\n", global, local, "failed to open window", allegro_error);
		return(1);
	}
	center_dx = (SCREEN_W - VIEW_WIDTH * scale_x) / 2;
	center_dy = (SCREEN_H - VIEW_HEIGHT * scale_y) / 2;

	set_window_title("X11 Emerald Mine");

	set_palette(privateColours);

	return(0);
}

/* handle frame sync, polling */

int keyboard_raw = -1;
int keyboard_cook = -1;

static int cursor_timer;
static int cursor_hidden;
static int cursor_motion;

static const int modifiers[12] = { /* these keycodes dont seem to be returned by readkey */
	KEY_LSHIFT, KEY_RSHIFT, KEY_LCONTROL, KEY_RCONTROL,
	KEY_ALT, KEY_ALTGR, KEY_LWIN, KEY_RWIN,
	KEY_MENU, KEY_SCRLOCK, KEY_NUMLOCK, KEY_CAPSLOCK,
};

static void audio_thread(void);

void display_read(void)
{
	int x;

	poll_keyboard();
	poll_joystick();
	poll_mouse();

	for(x = 0; x < 12; x++) {
		int kc = modifiers[x];
		if(key[kc]) { keyboard_raw = kc; keyboard_cook = -1; }
	}
	while(keypressed()) keyboard_cook = ureadkey(&keyboard_raw);
	if(keyboard_raw != -1 && !key[keyboard_raw]) { keyboard_raw = -1; keyboard_cook = -1; }

	x = cursor_motion;
	cursor_motion = mouse_pos;
	if(x != cursor_motion) cursor_timer = 32;
	if(cursor_timer == 0 && cursor_hidden == 0) {
		show_os_cursor(MOUSE_CURSOR_NONE);
		cursor_hidden = 1;
	}
	if(cursor_timer && cursor_hidden) {
		show_os_cursor(MOUSE_CURSOR_ARROW);
		cursor_hidden = 0;
	}
	if(cursor_timer) cursor_timer--;

	audio_thread();

	vsync();
}

int display_close(void)
{
	allegro_exit();
	return(0);
}

/* ---------------------------------------------------------------------- */

/* handle keyboard, key names */

char keyboard_input[2][INPUT_MAX];

int keyboard_keycode[2][INPUT_MAX];

static const int defaultKeyCode[INPUT_MAX][2] = {
	[INPUT_north] = { KEY_UP, KEY_W },
	[INPUT_east] = { KEY_RIGHT, KEY_D },
	[INPUT_south] = { KEY_DOWN, KEY_S },
	[INPUT_west] = { KEY_LEFT, KEY_A },
	[INPUT_shoot] = { KEY_RCONTROL, KEY_LCONTROL },
	[INPUT_abort] = { KEY_ESC, KEY_ESC },
};
static char keyboardString[KEY_MAX][12];

int keyboard_open(void)
{
	int ply, button, kc;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			keyboard_keycode[ply][button] = defaultKeyCode[button][ply];
		}
	}
	memset(keyboardString, 0, sizeof(keyboardString));
	for(kc = 0; kc < KEY_MAX; kc++) {
		ustrzcpy(keyboardString[kc], sizeof(keyboardString[kc]), scancode_to_name(kc));
	}
	for(kc = 0; kc < KEY_MAX; kc++) {
		char *ptr;
		for(ptr = keyboardString[kc]; *ptr; ptr++) if(*ptr == ' ') *ptr = '_';
	}
	return(0);
}

void keyboard_read(void)
{
	int ply, button;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			int kc = keyboard_keycode[ply][button] & 255;
			keyboard_input[ply][button] = key[kc];
		}
	}
}

int keyboard_close(void)
{
	return(0);
}

char *keyboard_to_name(int kc)
{
	return(keyboardString[kc & 255]);
}

int keyboard_to_keycode(char *kn)
{
	int kc = KEY_MAX;
	if(*kn) for(kc = 0; kc < KEY_MAX; kc++) if(ustrcmp(kn, keyboardString[kc]) == 0) break;
	if(kc >= KEY_MAX) kc = -1;
	return(kc);
}

/* ---------------------------------------------------------------------- */

/* handle joysticks */

char joystick_input[2][INPUT_MAX];

static int joy_fd[2] = { -1, -1 };

int joystick_open(void)
{
	int ply = 0;
	int i;
	for(i = 0; i < num_joysticks; i++) {
		if(ply < 2) joy_fd[ply++] = i;
	}
	return(0);
}

void joystick_read(void)
{
	int ply;
	for(ply = 0; ply < 2; ply++) {
		char *joyst = joystick_input[ply];
		int n = joy_fd[ply];
		if(n >= 0 && n < num_joysticks) {
			if(joy[n].num_sticks > 0) {
				if(joy[n].stick[0].num_axis > 0) {
					joyst[INPUT_west] = joy[n].stick[0].axis[0].d1;
					joyst[INPUT_east] = joy[n].stick[0].axis[0].d2;
				}
				if(joy[n].stick[0].num_axis > 1) {
					joyst[INPUT_north] = joy[n].stick[0].axis[1].d1;
					joyst[INPUT_south] = joy[n].stick[0].axis[1].d2;
				}
			}
			if(joy[n].num_buttons > 0) {
				joyst[INPUT_shoot] = joy[n].button[0].b;
			}
		}
	}
}

int joystick_close(void)
{
	joy_fd[0] = -1;
	joy_fd[1] = -1;
	return(0);
}

/* ---------------------------------------------------------------------- */

/* handle audio */

static char sound_play[SOUND_MAX]; /* should probably be protected by a semaphore */
static int sound_ptr[SOUND_MAX]; /* play position in the sound */

#define BUFFER_SIZE 256

static AUDIOSTREAM *audio_fd;

int audio_open(void)
{
	char buf[256];

	audio_fd = play_audio_stream(BUFFER_SIZE, 8, 0, 8000, 255, 128);
	if(audio_fd == 0) {
		uszprintf(buf, sizeof(buf), "%s[%s] %s", global, local, "failed to open audio device");
		goto fail;
	}
	memset(sound_ptr, -1, sizeof(sound_ptr));

	return(0);
fail:
	alert(buf, allegro_error, "", "ok", 0, '\r', '\n');
	return(1);
}

void audio_write(struct LOGIC *lev)
{
	int i;
	for(i = 0; i < SOUND_MAX; i++) if(lev->sound[i]) {
		sound_play[i] = 1;
	}
}

int audio_close(void)
{
	if(audio_fd) { stop_audio_stream(audio_fd); audio_fd = 0; }
	return(0);
}

/* mix audio data */

static int mix_play[MIXER_MAX]; /* sounds we have chosen to play */
static int mix_count;

static short buf_mix[BUFFER_SIZE]; /* buffer to mix sounds */

static void audio_mix(char *buf_out)
{
	int i;
	for(i = 0; i < mix_count; i++) {
		register const unsigned char *in = sounds_data[mix_play[i]] + sound_ptr[mix_play[i]];
		register short *out = buf_mix;
		register int len = sounds_length[mix_play[i]] - sound_ptr[mix_play[i]];
		if(len > BUFFER_SIZE) len = BUFFER_SIZE;
		do *out++ += ulaw_to_linear[*in++] / MIXER_MAX; while(--len);
	}
	{
		register const short *in = buf_mix;
		register char *out = buf_out;
		register int len = BUFFER_SIZE;
		do *out++ = (*in++ + 32768) >> 8; while(--len);
	}
}

/* audio callback */

static void audio_thread(void)
{
	char *buf_out; /* buffer written to device */
	int i;

	if(audio_fd == 0) return;

	for(; buf_out = get_audio_stream_buffer(audio_fd), buf_out; free_audio_stream_buffer(audio_fd)) {

		mix_count = 0;

		for(i = 0; i < SOUND_MAX; i++) if(sound_play[i]) {
			sound_play[i] = 0;
			sound_ptr[i] = 0;
		}
		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			while(mix_count < MIXER_MAX) {
				mix_play[mix_count++] = i;
				if(i >= SOUND_boom) break;
			}
		}

		memset(buf_mix, 0, sizeof(buf_mix));
		audio_mix(buf_out);

		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			sound_ptr[i] += BUFFER_SIZE;
			if(sound_ptr[i] >= sounds_length[i]) sound_ptr[i] = -1;
		}
	}
}

/* ---------------------------------------------------------------------- */

/* create bitmaps */

static BITMAP *screenPixmap;
static BITMAP *objectPixmap;
static BITMAP *scorePixmap;
static BITMAP *spritePixmap;

int screen_open(void)
{
	char buf[256];

	screenPixmap = create_bitmap((VIEW_WIDTH + SCROLL_WIDTH) * scale_x, (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y);
	if(screenPixmap == 0) {
		uszprintf(buf, sizeof(buf), "%s[%s] %s (%dx%dx%d)", global, local, "failed to create screen pixmap", (VIEW_WIDTH + SCROLL_WIDTH) * scale_x, (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y, get_color_depth());
		goto fail;
	}
	objectPixmap = create_bitmap(GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y);
	if(objectPixmap == 0) {
		uszprintf(buf, sizeof(buf), "%s[%s] %s (%dx%dx%d)", global, local, "failed to create graphics pixmap", GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, get_color_depth());
		goto fail;
	}
	scorePixmap = create_bitmap(SCORE_WIDTH * scale_x, 8 * scale_y);
	if(scorePixmap == 0) {
		uszprintf(buf, sizeof(buf), "%s[%s] %s (%dx%dx%d)", global, local, "failed to create score pixmap", SCORE_WIDTH * scale_x, 8 * scale_y, get_color_depth());
		goto fail;
	}
	spritePixmap = create_bitmap(16 * scale_x, 16 * scale_y);
	if(spritePixmap == 0) {
		uszprintf(buf, sizeof(buf), "%s[%s] %s (%dx%dx%d)", global, local, "failed to create sprite pixmap", 16 * scale_x, 16 * scale_y, get_color_depth());
		goto fail;
	}

	return(screen_write());
fail:
	alert(buf, allegro_error, "", "ok", 0, '\r', '\n');
	return(1);
}

/* render graphics */

int screen_write(void)
{
	int x, in_x, out_x;
	int y, in_y, out_y;

	out_y = 0;
	for(in_y = 0; in_y < GFX_HEIGHT; in_y++) {
		out_x = 0;
		for(in_x = 0; in_x < GFX_WIDTH; in_x++) {
			char c = graphics_xpm[in_x][in_y];
			unsigned long pixel = paletteHash[c & 127];
			for(x = 0; x < scale_x; x++) {
				putpixel(objectPixmap, out_x, out_y, pixel);
				out_x++;
			}
		}
		for(y = 1; y < scale_y; y++) {
			blit(objectPixmap, objectPixmap, 0, out_y, 0, out_y + y, GFX_WIDTH * scale_x, 1);
		}
		out_y += scale_y;
	}
	return(0);
}

int screen_close(void)
{
	if(screenPixmap) { destroy_bitmap(screenPixmap); screenPixmap = 0; }
	if(objectPixmap) { destroy_bitmap(objectPixmap); objectPixmap = 0; }
	if(scorePixmap) { destroy_bitmap(scorePixmap); scorePixmap = 0; }
	if(spritePixmap) { destroy_bitmap(spritePixmap); spritePixmap = 0; }
	return(0);
}

/* ---------------------------------------------------------------------- */

/* game drawing functions */

static short screen_cache[(VIEW_WIDTH + SCROLL_WIDTH) / 16][(VIEW_HEIGHT + SCROLL_HEIGHT) / 16];

static int scn_pixel_left;
static int scn_pixel_top;

void x11_game_clear(void)
{
	rectfill(screenPixmap, 0, 0, (VIEW_WIDTH + SCROLL_WIDTH) * scale_x - 1, (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y - 1, palette_color[0]);
	memset(screen_cache, -1, sizeof(screen_cache)); /* mark entire screen as dirty */
}

void x11_game_scroll(struct LOGIC *lev)
{
	scn_pixel_left = lev->pixel_left % SCROLL_WIDTH;
	scn_pixel_top = lev->pixel_top % SCROLL_HEIGHT;
}

void x11_game_object(struct LOGIC *lev)
{
	int tile, tile_x, tile_y;
	int scn_x, lev_x;
	int scn_y, lev_y;
	scn_x = scn_pixel_left / 16;
	for(lev_x = lev->view_left; lev_x <= lev->view_right; lev_x++) {
		scn_y = scn_pixel_top / 16;
		for(lev_y = lev->view_top; lev_y <= lev->view_bottom; lev_y++) {
			tile = object_map[lev->cave[lev_x][lev_y]][lev->frame];
			if(screen_cache[scn_x][scn_y] != tile) {
				screen_cache[scn_x][scn_y] = tile;
				tile_x = (tile / GFX_HEIGHT) * 16;
				tile_y = (tile % GFX_HEIGHT);
				blit(objectPixmap, screenPixmap, tile_x * scale_x, tile_y * scale_y, scn_x * 16 * scale_x, scn_y * 16 * scale_y, 16 * scale_x, 16 * scale_y);
			}
			scn_y++;
		}
		scn_x++;
	}
}

void x11_game_player(struct LOGIC *lev)
{
	int tile, tile_x, tile_y;
	int x0, x1;
	int y0, y1;
	int ply;
	for(ply = 0; ply < 2; ply++) {
		if(lev->hide[ply]) continue;

		x0 = lev->pl_pixel_left[ply];
		x1 = lev->pl_pixel_right[ply];
		y0 = lev->pl_pixel_top[ply];
		y1 = lev->pl_pixel_bottom[ply];
		if(!(x0 >= lev->pixel_left && x1 <= lev->pixel_right && y0 >= lev->pixel_top && y1 <= lev->pixel_bottom)) continue;

		x0 -= lev->pixel_left - scn_pixel_left;
		x1 -= lev->pixel_left - scn_pixel_left;
		y0 -= lev->pixel_top - scn_pixel_top;
		y1 -= lev->pixel_top - scn_pixel_top;

		screen_cache[x0 / 16][y0 / 16] = -1; /* mark screen over player as dirty */
		screen_cache[x1 / 16][y1 / 16] = -1;

		tile = sprite_map[lev->pl_anim[ply]][ply][lev->frame];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		blit(objectPixmap, spritePixmap, tile_x * scale_x, tile_y * scale_y, 0, 0, 16 * scale_x, 16 * scale_y);
		masked_blit(screenPixmap, spritePixmap, x0 * scale_x, y0 * scale_y, 0, 0, 16 * scale_x, 16 * scale_y);
		blit(spritePixmap, screenPixmap, 0, 0, x0 * scale_x, y0 * scale_y, 16 * scale_x, 16 * scale_y);
	}
}

void x11_game_score(struct LOGIC *lev)
{
	int x = scn_pixel_left, x0, x1;
	int y = scn_pixel_top + 4, y0, y1;

	masked_blit(scorePixmap, screenPixmap, 0, 0, x * scale_x, y * scale_y, SCORE_WIDTH * scale_x, 8 * scale_y);

	x0 = x / 16; x1 = (x + SCORE_WIDTH - 1) / 16;
	y0 = y / 16; y1 = (y + 8 - 1) / 16;
	for(x = x0; x <= x1; x++) for(y = y0; y <= y1; y++) {
		screen_cache[x][y] = -1; /* mark screen under score as dirty */
	}
}

void x11_game_to_screen(void)
{
	blit(screenPixmap, screen, scn_pixel_left * scale_x, scn_pixel_top * scale_y, center_dx, center_dy, VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y);
}

/* ---------------------------------------------------------------------- */

/* score drawing functions */

static short score_cache[SCORE_WIDTH / 8];

void x11_score_clear(void)
{
	rectfill(scorePixmap, 0, 0, SCORE_WIDTH * scale_x - 1, 8 * scale_y - 1, palette_color[0]);
	memset(score_cache, -1, sizeof(score_cache));
}

void x11_score_print(char str[40])
{
	int tile, tile_x, tile_y;
	int x;
	for(x = 0; x < SCORE_WIDTH / 8; x++) {
		tile = score_map[str[x] & 127];
		if(score_cache[x] != tile) {
			score_cache[x] = tile;
			tile_x = (tile / GFX_HEIGHT) * 16;
			tile_y = (tile % GFX_HEIGHT);
			blit(objectPixmap, scorePixmap, tile_x * scale_x, tile_y * scale_y, x * 8 * scale_x, 0, 8 * scale_x, 8 * scale_y);
		}
	}
}

/* ---------------------------------------------------------------------- */

/* menu drawing functions */

static int antsPos;

void x11_menu_clear(void)
{
	rectfill(screenPixmap, 0, 0, VIEW_WIDTH * scale_x - 1, VIEW_HEIGHT * scale_y - 1, palette_color[60 + 1]);
}

void x11_menu_print(char str[26], int top)
{
	int tile, tile_x, tile_y;
	int x;
	rectfill(screenPixmap, 0, top * 16 * scale_y, VIEW_WIDTH * scale_x - 1, (top + 1) * 16 * scale_y - 1, palette_color[60 + 1]);
	for(x = 0; x < 26; x++) {
		tile = font_map[str[x] & 127];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		masked_blit(objectPixmap, screenPixmap, tile_x * scale_x, tile_y * scale_y, (x * 12 + 4) * scale_x, top * 16 * scale_y, 16 * scale_x, 16 * scale_y);
	}
}

void x11_menu_ants(int left, int right, int top)
{
	unsigned long whitePixel = palette_color[COLOUR_white + 1];
	unsigned long redPixel = palette_color[COLOUR_red + 1];
	int x, x0 = left, x1 = (right + 1);
	int y, y0 = top * 16, y1 = (top + 1) * 16;
	int pos, inc;

	antsPos = (antsPos + 1) % 16;
	pos = antsPos;

	for(x = x0; x < x1; x += inc) {
		inc = 8 - pos % 8; if(inc > x1 - x) inc = x1 - x;
		rectfill(screenPixmap, x * scale_x, y0 * scale_y, (x + inc) * scale_x - 1, (y0 + 1) * scale_y - 1, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(y = y0; y < y1; y += inc) {
		inc = 8 - pos % 8; if(inc > y1 - y) inc = y1 - y;
		rectfill(screenPixmap, (x1 - 1) * scale_x, y * scale_y, x1 * scale_x - 1, (y + inc) * scale_y - 1, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(x = x1; x > x0; x -= inc) {
		inc = 8 - pos % 8; if(inc > x - x0) inc = x - x0;
		rectfill(screenPixmap, (x - inc) * scale_x, (y1 - 1) * scale_y, x * scale_x - 1, y1 * scale_y - 1, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(y = y1; y > y0; y -= inc) {
		inc = 8 - pos % 8; if(inc > y - y0) inc = y - y0;
		rectfill(screenPixmap, x0 * scale_x, (y - inc) * scale_y, (x0 + 1) * scale_x - 1, y * scale_y - 1, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
}

void x11_menu_to_screen(void)
{
	blit(screenPixmap, screen, 0, 0, center_dx, center_dy, VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y);
}
