/* handle display, keyboard, joystick, and audio in emerald mine */

/* uses simple directmedia layer */

extern char *global;
static char *local = "sdl";

#include "proto.h"

#include <SDL/SDL.h>

#include <stdio.h>

/* ---------------------------------------------------------------------- */

static void parseGeometry(char *parsestring, int *x_return, int *y_return, unsigned int *width_return, unsigned int *height_return)
{
	int x = 0, y = 0;
	unsigned int width = 0, height = 0;
	char gotxy = 0, gotwh = 0;
	char *endptr;
	if(*parsestring == '=') parsestring++;
	if(*parsestring != '+' && *parsestring != '-') {
		width = SDL_strtoul(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		if(*parsestring != 'x' && *parsestring != 'X') return;
		parsestring++;
		height = SDL_strtoul(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		gotwh = 1;
	}
	if(*parsestring == '+' || *parsestring == '-') {
		parsestring++;
		x = SDL_strtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		if(*parsestring != '+' && *parsestring != '-') return;
		parsestring++;
		y = SDL_strtol(parsestring, &endptr, 10);
		if(parsestring == endptr) return;
		parsestring = endptr;
		gotxy = 1;
	}
	if(*parsestring != 0) return;

	if(gotxy) {
		if(x_return) *x_return = x;
		if(y_return) *y_return = y;
	}
	if(gotwh) {
		if(width_return) *width_return = width;
		if(height_return) *height_return = height;
	}
}

/* ---------------------------------------------------------------------- */

/* initialize sdl, create window, create palette */

static int scale_x = 2; /* default to double size because it looks better on most displays */
static int scale_y = 2;

static int center_dx, center_dy;

static SDL_Surface *window;

static SDL_Color privateColours[COLOUR_MAX + 1];
static unsigned long paletteHash[128];

int display_open(void)
{
	int i;

	if(SDL_Init(SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_NOPARACHUTE) == -1) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to initialize system", SDL_GetError());
		return(1);
	}

	for(i = 0; i < COLOUR_MAX; i++) {
		unsigned long rgbc = colours_xpm[i];
		privateColours[i + 1].r = (rgbc >> 24) & 255;
		privateColours[i + 1].g = (rgbc >> 16) & 255;
		privateColours[i + 1].b = (rgbc >> 8) & 255;
		paletteHash[rgbc & 127] = i + 1;
	}

	if(arg_geometry) {
		int dummy;
		unsigned int w = scale_x;
		unsigned int h = scale_y;
		parseGeometry(arg_geometry, &dummy, &dummy, &w, &h);
		scale_x = w;
		scale_y = h;
	}
	if(scale_x < 1 || scale_x > 9 || scale_y < 1 || scale_y > 9) {
		scale_x = 1;
		scale_y = 1;
	}

	if(SDL_SetVideoMode(VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y, 0, SDL_SWSURFACE | SDL_ANYFORMAT | (arg_install ? SDL_HWPALETTE : 0) | (arg_fullscreen ? SDL_FULLSCREEN : 0)) == 0) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to open window", SDL_GetError());
		return(1);
	}
	window = SDL_GetVideoSurface();
	center_dx = (window->w - VIEW_WIDTH * scale_x) / 2;
	center_dy = (window->h - VIEW_HEIGHT * scale_y) / 2;

	SDL_WM_SetCaption("X11 Emerald Mine", global);
	SDL_EnableUNICODE(1);

	SDL_SetPalette(window, SDL_LOGPAL | (arg_install ? SDL_PHYSPAL : 0), privateColours, 0, COLOUR_MAX + 1);

	return(0);
}

/* handle frame sync, events */

int keyboard_raw = -1;
int keyboard_cook = -1;

static long frame_msec;

static int cursor_timer;
static int cursor_hidden;

void display_read(void)
{
	long msec;
	SDL_Event event;

	SDL_JoystickUpdate();

	while(SDL_PollEvent(&event)) {
		switch(event.type) {
		case SDL_KEYDOWN:
			keyboard_raw = event.key.keysym.sym;
			keyboard_cook = event.key.keysym.unicode;
			break;
		case SDL_KEYUP:
			keyboard_raw = -1;
			keyboard_cook = -1;
			break;
		case SDL_MOUSEMOTION:
			cursor_timer = 32;
			break;
		}
	}
	if(cursor_timer == 0 && cursor_hidden == 0) {
		SDL_ShowCursor(SDL_DISABLE);
		cursor_hidden = 1;
	}
	if(cursor_timer && cursor_hidden) {
		SDL_ShowCursor(SDL_ENABLE);
		cursor_hidden = 0;
	}
	if(cursor_timer) cursor_timer--;

	msec = frame_msec + 1000;
	frame_msec = SDL_GetTicks() * 60;
	msec -= frame_msec;
	if(msec >= 0 && msec <= 1500) {
		frame_msec += msec;
		SDL_Delay(msec / 60);
	}

	SDL_Flip(window);
}

int display_close(void)
{
	SDL_Quit();
	return(0);
}

/* ---------------------------------------------------------------------- */

/* handle keyboard, key names */

char keyboard_input[2][INPUT_MAX];

int keyboard_keycode[2][INPUT_MAX];

static const int defaultKeyCode[INPUT_MAX][2] = {
	[INPUT_north] = { SDLK_UP, SDLK_w },
	[INPUT_east] = { SDLK_RIGHT, SDLK_d },
	[INPUT_south] = { SDLK_DOWN, SDLK_s },
	[INPUT_west] = { SDLK_LEFT, SDLK_a },
	[INPUT_shoot] = { SDLK_RCTRL, SDLK_LCTRL },
	[INPUT_abort] = { SDLK_ESCAPE, SDLK_ESCAPE },
};
static Uint8 *keyboardVector;

static char keyboardString[512][12];

int keyboard_open(void)
{
	int ply, button, kc;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			keyboard_keycode[ply][button] = defaultKeyCode[button][ply];
		}
	}
	SDL_memset(keyboardString, 0, sizeof(keyboardString));
	for(kc = 0; kc < 512; kc++) {
		SDL_strlcpy(keyboardString[kc], SDL_GetKeyName(kc), sizeof(keyboardString[kc]));
	}
	for(kc = 0; kc < 512; kc++) {
		char *ptr;
		for(ptr = keyboardString[kc]; *ptr; ptr++) if(*ptr == ' ') *ptr = '_';
	}
	keyboardVector = SDL_GetKeyState(0);
	return(0);
}

void keyboard_read(void)
{
	int ply, button;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			int kc = keyboard_keycode[ply][button] & 511;
			keyboard_input[ply][button] = keyboardVector[kc];
		}
	}
}

int keyboard_close(void)
{
	return(0);
}

char *keyboard_to_name(int kc)
{
	return(keyboardString[kc & 511]);
}

int keyboard_to_keycode(char *kn)
{
	int kc = 512;
	if(*kn) for(kc = 0; kc < 512; kc++) if(SDL_strcmp(kn, keyboardString[kc]) == 0) break;
	if(kc >= 512) kc = -1;
	return(kc);
}

/* ---------------------------------------------------------------------- */

/* handle joysticks */

char joystick_input[2][INPUT_MAX];

static SDL_Joystick *joy_fd[2];

int joystick_open(void)
{
	int ply = 0;
	int i;
	for(i = 0; i < SDL_NumJoysticks(); i++) {
		if(ply < 2) {
			SDL_Joystick *fd = SDL_JoystickOpen(i);
			if(fd == 0) {
				fprintf(stderr, "%s[%s] %s: %s: %s\n", global, local, SDL_JoystickName(i), "failed to open joystick", SDL_GetError());
				continue;
			}
			joy_fd[ply++] = fd;
		}
	}
	return(0);
}

void joystick_read(void)
{
	int ply;

	for(ply = 0; ply < 2; ply++) {
		char *joy = joystick_input[ply];

		int xpos = SDL_JoystickGetAxis(joy_fd[ply], 0);
		int ypos = SDL_JoystickGetAxis(joy_fd[ply], 1);
		int shoot = SDL_JoystickGetButton(joy_fd[ply], 0);

		joy[INPUT_north] = (ypos < -0x4000);
		joy[INPUT_east] = (xpos > 0x4000);
		joy[INPUT_south] = (ypos > 0x4000);
		joy[INPUT_west] = (xpos < -0x4000);
		joy[INPUT_shoot] = (shoot != 0);
	}
}

int joystick_close(void)
{
	int ply;
	for(ply = 0; ply < 2; ply++) {
		if(joy_fd[ply]) { SDL_JoystickClose(joy_fd[ply]); joy_fd[ply] = 0; }
	}
	return(0);
}

/* ---------------------------------------------------------------------- */

/* handle audio */

static char sound_play[SOUND_MAX]; /* should probably be protected by a semaphore */
static int sound_ptr[SOUND_MAX]; /* play position in the sound */

#define BUFFER_SIZE 256

static SDL_AudioSpec audioInfo;

static void audio_thread(void *userdata, Uint8 *stream, int length);

int audio_open(void)
{
	SDL_memset(sound_ptr, -1, sizeof(sound_ptr));

	audioInfo.freq = 8000;
	audioInfo.format = AUDIO_U8;
	audioInfo.channels = 1;
	audioInfo.silence = 128;
	audioInfo.samples = BUFFER_SIZE;
	audioInfo.size = BUFFER_SIZE;
	audioInfo.callback = audio_thread;

	if(SDL_OpenAudio(&audioInfo, 0) == -1) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to open audio device", SDL_GetError());
		return(1);
	}
	SDL_PauseAudio(0);
	return(0);
}

void audio_write(struct LOGIC *lev)
{
	int i;
	for(i = 0; i < SOUND_MAX; i++) if(lev->sound[i]) {
		sound_play[i] = 1;
	}
}

int audio_close(void)
{
	SDL_CloseAudio();
	return(0);
}

/* mix audio data */

static int mix_play[MIXER_MAX]; /* sounds we have chosen to play */
static int mix_count;

static short buf_mix[BUFFER_SIZE]; /* buffer to mix sounds */

static void audio_mix(Uint8 *stream, int length)
{
	int i;
	for(i = 0; i < mix_count; i++) {
		register const unsigned char *in = sounds_data[mix_play[i]] + sound_ptr[mix_play[i]];
		register short *out = buf_mix;
		register int len = sounds_length[mix_play[i]] - sound_ptr[mix_play[i]];
		if(len > BUFFER_SIZE) len = BUFFER_SIZE;
		do *out++ += ulaw_to_linear[*in++] / MIXER_MAX; while(--len);
	}
	{
		register const short *in = buf_mix;
		register Uint8 *out = stream;
		register int len = BUFFER_SIZE;
		if(len > length) len = length;
		do *out++ = (*in++ + 32768) >> 8; while(--len);
	}
}

/* audio callback */

static void audio_thread(void *userdata, Uint8 *stream, int length)
{
	int i;

	mix_count = 0;

	for(i = 0; i < SOUND_MAX; i++) if(sound_play[i]) {
		sound_play[i] = 0;
		sound_ptr[i] = 0;
	}
	for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
		while(mix_count < MIXER_MAX) {
			mix_play[mix_count++] = i;
			if(i >= SOUND_boom) break;
		}
	}

	SDL_memset(buf_mix, 0, sizeof(buf_mix));
	audio_mix(stream, length);

	for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
		sound_ptr[i] += BUFFER_SIZE;
		if(sound_ptr[i] >= sounds_length[i]) sound_ptr[i] = -1;
	}
}

/* ---------------------------------------------------------------------- */

/* create surfaces */

static SDL_Surface *screenPixmap;
static SDL_Surface *objectPixmap;
static SDL_Surface *scorePixmap;
static SDL_Surface *spritePixmap;

int screen_open(void)
{
	screenPixmap = SDL_CreateRGBSurface(SDL_SWSURFACE, (VIEW_WIDTH + SCROLL_WIDTH) * scale_x, (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y, 8, 0, 0, 0, 0);
	if(screenPixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, "failed to create screen pixmap", SDL_GetError(), (VIEW_WIDTH + SCROLL_WIDTH) * scale_x, (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y, 8);
		return(1);
	}
	objectPixmap = SDL_CreateRGBSurface(SDL_SWSURFACE, GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, 8, 0, 0, 0, 0);
	if(objectPixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, "failed to create graphics pixmap", SDL_GetError(), GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, 8);
		return(1);
	}
	scorePixmap = SDL_CreateRGBSurface(SDL_SWSURFACE, SCORE_WIDTH * scale_x, 8 * scale_y, 8, 0, 0, 0, 0);
	if(scorePixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, "failed to create score pixmap", SDL_GetError(), SCORE_WIDTH * scale_x, 8 * scale_y, 8);
		return(1);
	}
	spritePixmap = SDL_CreateRGBSurface(SDL_SWSURFACE, 16 * scale_x, 16 * scale_y, 8, 0, 0, 0, 0);
	if(spritePixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, "failed to create sprite pixmap", SDL_GetError(), 16 * scale_x, 16 * scale_y, 8);
		return(1);
	}
	SDL_SetColors(screenPixmap, privateColours, 0, COLOUR_MAX + 1);
	SDL_SetColors(objectPixmap, privateColours, 0, COLOUR_MAX + 1);
	SDL_SetColors(scorePixmap, privateColours, 0, COLOUR_MAX + 1);
	SDL_SetColors(spritePixmap, privateColours, 0, COLOUR_MAX + 1);

	return(screen_write());
}

/* render graphics */

int screen_write(void)
{
	int x, in_x, out_x;
	int y, in_y, out_y;

	SDL_LockSurface(objectPixmap);
	out_y = 0;
	for(in_y = 0; in_y < GFX_HEIGHT; in_y++) {
		Uint8 *pixmap_data = objectPixmap->pixels;
		Uint8 *pixmap_ptr = pixmap_data + out_y * objectPixmap->pitch;
		out_x = 0;
		for(in_x = 0; in_x < GFX_WIDTH; in_x++) {
			char c = graphics_xpm[in_x][in_y];
			unsigned long pixel = paletteHash[c & 127];
			for(x = 0; x < scale_x; x++) {
				pixmap_ptr[out_x] = pixel;
				out_x++;
			}
		}
		for(y = 1; y < scale_y; y++) {
			SDL_memcpy(pixmap_ptr + y * objectPixmap->pitch, pixmap_ptr, GFX_WIDTH * scale_x);
		}
		out_y += scale_y;
	}
	SDL_UnlockSurface(objectPixmap);

	return(0);
}

int screen_close(void)
{
	if(screenPixmap) { SDL_FreeSurface(screenPixmap); screenPixmap = 0; }
	if(objectPixmap) { SDL_FreeSurface(objectPixmap); objectPixmap = 0; }
	if(scorePixmap) { SDL_FreeSurface(scorePixmap); scorePixmap = 0; }
	if(spritePixmap) { SDL_FreeSurface(spritePixmap); spritePixmap = 0; }
	return(0);
}

/* ---------------------------------------------------------------------- */

/* game drawing functions */

static short screen_cache[(VIEW_WIDTH + SCROLL_WIDTH) / 16][(VIEW_HEIGHT + SCROLL_HEIGHT) / 16];

static int scn_pixel_left;
static int scn_pixel_top;

void x11_game_clear(void)
{
	SDL_Rect r;
	r.x = 0; r.w = (VIEW_WIDTH + SCROLL_WIDTH) * scale_x;
	r.y = 0; r.h = (VIEW_HEIGHT + SCROLL_HEIGHT) * scale_y;
	SDL_FillRect(screenPixmap, &r, 0);
	SDL_memset(screen_cache, -1, sizeof(screen_cache)); /* mark entire screen as dirty */
}

void x11_game_scroll(struct LOGIC *lev)
{
	scn_pixel_left = lev->pixel_left % SCROLL_WIDTH;
	scn_pixel_top = lev->pixel_top % SCROLL_HEIGHT;
}

void x11_game_object(struct LOGIC *lev)
{
	SDL_Rect srcR, dstR;
	int tile;
	int scn_x, lev_x;
	int scn_y, lev_y;
	scn_x = scn_pixel_left / 16;
	for(lev_x = lev->view_left; lev_x <= lev->view_right; lev_x++) {
		scn_y = scn_pixel_top / 16;
		for(lev_y = lev->view_top; lev_y <= lev->view_bottom; lev_y++) {
			tile = object_map[lev->cave[lev_x][lev_y]][lev->frame];
			if(screen_cache[scn_x][scn_y] != tile) {
				screen_cache[scn_x][scn_y] = tile;
				srcR.x = (tile / GFX_HEIGHT) * 16 * scale_x; dstR.x = scn_x * 16 * scale_x; srcR.w = 16 * scale_x;
				srcR.y = (tile % GFX_HEIGHT) * scale_y; dstR.y = scn_y * 16 * scale_y; srcR.h = 16 * scale_y;
				SDL_BlitSurface(objectPixmap, &srcR, screenPixmap, &dstR);
			}
			scn_y++;
		}
		scn_x++;
	}
}

void x11_game_player(struct LOGIC *lev)
{
	SDL_Rect srcR, dstR;
	int tile;
	int x0, x1;
	int y0, y1;
	int ply;
	for(ply = 0; ply < 2; ply++) {
		if(lev->hide[ply]) continue;

		x0 = lev->pl_pixel_left[ply];
		x1 = lev->pl_pixel_right[ply];
		y0 = lev->pl_pixel_top[ply];
		y1 = lev->pl_pixel_bottom[ply];
		if(!(x0 >= lev->pixel_left && x1 <= lev->pixel_right && y0 >= lev->pixel_top && y1 <= lev->pixel_bottom)) continue;

		x0 -= lev->pixel_left - scn_pixel_left;
		x1 -= lev->pixel_left - scn_pixel_left;
		y0 -= lev->pixel_top - scn_pixel_top;
		y1 -= lev->pixel_top - scn_pixel_top;

		screen_cache[x0 / 16][y0 / 16] = -1; /* mark screen over player as dirty */
		screen_cache[x1 / 16][y1 / 16] = -1;

		tile = sprite_map[lev->pl_anim[ply]][ply][lev->frame];
		srcR.x = (tile / GFX_HEIGHT) * 16 * scale_x; dstR.x = 0; srcR.w = 16 * scale_x;
		srcR.y = (tile % GFX_HEIGHT) * scale_y; dstR.y = 0; srcR.h = 16 * scale_y;
		SDL_BlitSurface(objectPixmap, &srcR, spritePixmap, &dstR);
		srcR.x = x0 * scale_x; dstR.x = 0;
		srcR.y = y0 * scale_y; dstR.y = 0;
		SDL_SetColorKey(screenPixmap, SDL_SRCCOLORKEY, 0);
		SDL_BlitSurface(screenPixmap, &srcR, spritePixmap, &dstR);
		SDL_SetColorKey(screenPixmap, 0, 0);
		srcR.x = 0; dstR.x = x0 * scale_x;
		srcR.y = 0; dstR.y = y0 * scale_y;
		SDL_BlitSurface(spritePixmap, &srcR, screenPixmap, &dstR);
	}
}

void x11_game_score(struct LOGIC *lev)
{
	SDL_Rect srcR, dstR;
	int x = scn_pixel_left, x0, x1;
	int y = scn_pixel_top + 4, y0, y1;

	srcR.x = 0; dstR.x = x * scale_x; srcR.w = SCORE_WIDTH * scale_x;
	srcR.y = 0; dstR.y = y * scale_y; srcR.h = 8 * scale_y;
	SDL_SetColorKey(scorePixmap, SDL_SRCCOLORKEY, 0);
	SDL_BlitSurface(scorePixmap, &srcR, screenPixmap, &dstR);
	SDL_SetColorKey(scorePixmap, 0, 0);

	x0 = x / 16; x1 = (x + SCORE_WIDTH - 1) / 16;
	y0 = y / 16; y1 = (y + 8 - 1) / 16;
	for(x = x0; x <= x1; x++) for(y = y0; y <= y1; y++) {
		screen_cache[x][y] = -1; /* mark screen under score as dirty */
	}
}

void x11_game_to_screen(void)
{
	SDL_Rect srcR, dstR;
	srcR.x = scn_pixel_left * scale_x; dstR.x = center_dx; srcR.w = VIEW_WIDTH * scale_x;
	srcR.y = scn_pixel_top * scale_y; dstR.y = center_dy; srcR.h = VIEW_HEIGHT * scale_y;
	SDL_BlitSurface(screenPixmap, &srcR, window, &dstR);
}

/* ---------------------------------------------------------------------- */

/* score drawing functions */

static short score_cache[SCORE_WIDTH / 8];

void x11_score_clear(void)
{
	SDL_Rect r;
	r.x = 0; r.w = SCORE_WIDTH * scale_x;
	r.y = 0; r.h = 8 * scale_y;
	SDL_FillRect(scorePixmap, &r, 0);
	SDL_memset(score_cache, -1, sizeof(score_cache));
}

void x11_score_print(char str[40])
{
	SDL_Rect srcR, dstR;
	int tile;
	int x;
	for(x = 0; x < SCORE_WIDTH / 8; x++) {
		tile = score_map[str[x] & 127];
		if(score_cache[x] != tile) {
			score_cache[x] = tile;
			srcR.x = (tile / GFX_HEIGHT) * 16 * scale_x; dstR.x = x * 8 * scale_x; srcR.w = 8 * scale_x;
			srcR.y = (tile % GFX_HEIGHT) * scale_y; dstR.y = 0; srcR.h = 8 * scale_y;
			SDL_BlitSurface(objectPixmap, &srcR, scorePixmap, &dstR);
		}
	}
}

/* ---------------------------------------------------------------------- */

/* menu drawing functions */

static int antsPos;

void x11_menu_clear(void)
{
	Uint32 blackPixel = SDL_MapRGB(screenPixmap->format, privateColours[60 + 1].r, privateColours[60 + 1].g, privateColours[60 + 1].b);
	SDL_Rect r;
	r.x = 0; r.w = VIEW_WIDTH * scale_x;
	r.y = 0; r.h = VIEW_HEIGHT * scale_y;
	SDL_FillRect(screenPixmap, &r, blackPixel);
}

void x11_menu_print(char str[26], int top)
{
	Uint32 blackPixel = SDL_MapRGB(screenPixmap->format, privateColours[60 + 1].r, privateColours[60 + 1].g, privateColours[60 + 1].b);
	SDL_Rect srcR, dstR;
	int tile;
	int x;
	srcR.x = 0; srcR.w = VIEW_WIDTH * scale_x;
	srcR.y = top * 16 * scale_y; srcR.h = 16 * scale_y;
	SDL_FillRect(screenPixmap, &srcR, blackPixel);
	SDL_SetColorKey(objectPixmap, SDL_SRCCOLORKEY, 0);
	for(x = 0; x < 26; x++) {
		tile = font_map[str[x] & 127];
		srcR.x = (tile / GFX_HEIGHT) * 16 * scale_x; dstR.x = (x * 12 + 4) * scale_x; srcR.w = 16 * scale_x;
		srcR.y = (tile % GFX_HEIGHT) * scale_y; dstR.y = top * 16 * scale_y; srcR.h = 16 * scale_y;
		SDL_BlitSurface(objectPixmap, &srcR, screenPixmap, &dstR);
	}
	SDL_SetColorKey(objectPixmap, 0, 0);
}

void x11_menu_ants(int left, int right, int top)
{
	Uint32 whitePixel = SDL_MapRGB(screenPixmap->format, privateColours[COLOUR_white + 1].r, privateColours[COLOUR_white + 1].g, privateColours[COLOUR_white + 1].b);
	Uint32 redPixel = SDL_MapRGB(screenPixmap->format, privateColours[COLOUR_red + 1].r, privateColours[COLOUR_red + 1].g, privateColours[COLOUR_red + 1].b);
	SDL_Rect r;
	int x, x0 = left, x1 = (right + 1);
	int y, y0 = top * 16, y1 = (top + 1) * 16;
	int pos, inc;

	antsPos = (antsPos + 1) % 16;
	pos = antsPos;

	for(x = x0; x < x1; x += inc) {
		inc = 8 - pos % 8; if(inc > x1 - x) inc = x1 - x;
		r.x = x * scale_x; r.w = inc * scale_x;
		r.y = y0 * scale_y; r.h = 1 * scale_y;
		SDL_FillRect(screenPixmap, &r, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(y = y0; y < y1; y += inc) {
		inc = 8 - pos % 8; if(inc > y1 - y) inc = y1 - y;
		r.x = (x1 - 1) * scale_x; r.w = 1 * scale_x;
		r.y = y * scale_y; r.h = inc * scale_y;
		SDL_FillRect(screenPixmap, &r, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(x = x1; x > x0; x -= inc) {
		inc = 8 - pos % 8; if(inc > x - x0) inc = x - x0;
		r.x = (x - inc) * scale_x; r.w = inc * scale_x;
		r.y = (y1 - 1) * scale_y; r.h = 1 * scale_y;
		SDL_FillRect(screenPixmap, &r, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
	for(y = y1; y > y0; y -= inc) {
		inc = 8 - pos % 8; if(inc > y - y0) inc = y - y0;
		r.x = x0 * scale_x; r.w = 1 * scale_x;
		r.y = (y - inc) * scale_y; r.h = inc * scale_y;
		SDL_FillRect(screenPixmap, &r, pos / 8 % 2 ? whitePixel : redPixel);
		pos += inc;
	}
}

void x11_menu_to_screen(void)
{
	SDL_Rect srcR, dstR;
	srcR.x = 0; dstR.x = center_dx; srcR.w = VIEW_WIDTH * scale_x;
	srcR.y = 0; dstR.y = center_dy; srcR.h = VIEW_HEIGHT * scale_y;
	SDL_BlitSurface(screenPixmap, &srcR, window, &dstR);
}
