/* handle display, keyboard, joystick, and audio in emerald mine */

/* uses unix X11 */

extern char *global;
static char *local = "x11";

#include "proto.h"

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>

#include "Icon.xbm"

/* ---------------------------------------------------------------------- */

/* open display, create window, allocate colours, create icon, allocate cursor */

static int scale_x = 2; /* default to double size because it looks better on most displays */
static int scale_y = 2;

static int center_dx, center_dy;

static char *displayName;
static Display *display;
static Window window;

static XVisualInfo visualInfo;

static unsigned long blackPixel;
static Colormap defaultColourmap;
static XColor allocColour;

static Colormap privateColourmap;
static int privateNumColours;
static XColor *privateColours;

static int paletteNumUsed;
static unsigned long palettePixels[COLOUR_MAX];
static unsigned long paletteHash[128];

static XSizeHints sizeHints;
static XSetWindowAttributes setWindowAttributes;
static XWMHints wmHints;
static Atom deleteAtom;

static Pixmap cursorBitmap;
static Cursor cursor;

static unsigned long privateAllocColour(unsigned short red, unsigned short green, unsigned short blue);

int display_open(void)
{
	int i;

	displayName = XDisplayName(arg_display);
	display = XOpenDisplay(arg_display);
	if(display == 0) {
		fprintf(stderr, "%s[%s] %s: %s: %s\n", global, local, displayName, "failed to open display", strerror(errno));
		return(1);
	}

	visualInfo.screen = DefaultScreen(display);
	visualInfo.visual = DefaultVisual(display, visualInfo.screen);
	visualInfo.depth = DefaultDepth(display, visualInfo.screen);
	visualInfo.class = PseudoColor;

	blackPixel = BlackPixel(display, visualInfo.screen);
	defaultColourmap = DefaultColormap(display, visualInfo.screen);
	allocColour.flags = DoRed | DoGreen | DoBlue;

	if(arg_install) {
		XVisualInfo *visualptr = XGetVisualInfo(display, VisualScreenMask | VisualClassMask, &visualInfo, &i);
		if(visualptr == 0) {
			fprintf(stderr, "%s[%s] %s: %s\n", global, local, displayName, "failed to find a pseudocolour visual");
			return(1);
		}
		while(i > 0) if(visualptr[--i].depth == 8) break;
		memcpy(&visualInfo, visualptr + i, sizeof(visualInfo));
		XFree(visualptr);

		blackPixel = COLOUR_black;

		privateColourmap = XCreateColormap(display, RootWindow(display, visualInfo.screen), visualInfo.visual, AllocAll);
		if(privateColourmap == 0) {
			fprintf(stderr, "%s[%s] %s: %s\n", global, local, displayName, "failed to create colourmap");
			return(1);
		}
		privateNumColours = visualInfo.colormap_size;

		privateColours = malloc(privateNumColours * sizeof(XColor));
		if(privateColours == 0) {
			fprintf(stderr, "%s[%s] %s: %s (%d)\n", global, local, displayName, "failed to allocate memory for colours", privateNumColours * sizeof(XColor));
			return(1);
		}
		for(i = 0; i < privateNumColours; i++) privateColours[i].pixel = i;
		XQueryColors(display, defaultColourmap, privateColours, privateNumColours);
		privateColours[blackPixel].red = 0; privateColours[blackPixel].green = 0; privateColours[blackPixel].blue = 0;
		XStoreColors(display, privateColourmap, privateColours, privateNumColours);
		privateColours[blackPixel].flags = 0;
	}

	for(i = 0; i < 128; i++) paletteHash[i] = blackPixel;

	for(paletteNumUsed = 0; paletteNumUsed < COLOUR_MAX; paletteNumUsed++) {
		unsigned long rgbc = colours_xpm[paletteNumUsed];
		allocColour.red = ((rgbc >> 24) & 255) * 257;
		allocColour.green = ((rgbc >> 16) & 255) * 257;
		allocColour.blue = ((rgbc >> 8) & 255) * 257;
		if(privateColourmap == 0) {
			if(XAllocColor(display, defaultColourmap, &allocColour) == 0) {
				fprintf(stderr, "%s[%s] %s: %s: %d (#%04x%04x%04x)\n", global, local, displayName, "failed to allocate colour", paletteNumUsed, allocColour.red, allocColour.green, allocColour.blue);
				return(1);
			}
		} else {
			allocColour.pixel = privateAllocColour(allocColour.red, allocColour.green, allocColour.blue);
		}
		palettePixels[paletteNumUsed] = allocColour.pixel;
		paletteHash[rgbc & 127] = allocColour.pixel;
	}

	sizeHints.x = 0;
	sizeHints.y = 0;
	if(arg_fullscreen) {
		sizeHints.width = DisplayWidth(display, visualInfo.screen);
		sizeHints.height = DisplayHeight(display, visualInfo.screen);
		scale_x = sizeHints.width / VIEW_WIDTH;
		scale_y = sizeHints.height / VIEW_HEIGHT;
		if(scale_x < scale_y) scale_y = scale_x; else scale_x = scale_y; /* keep it square */
	}
	if(arg_geometry) {
		int dummy;
		unsigned int w = scale_x;
		unsigned int h = scale_y;
		XParseGeometry(arg_geometry, &dummy, &dummy, &w, &h);
		scale_x = w;
		scale_y = h;
	}
	if(scale_x < 1 || scale_x > 9 || scale_y < 1 || scale_y > 9) {
		scale_x = 1;
		scale_y = 1;
	}
	if(!arg_fullscreen) {
		sizeHints.width = VIEW_WIDTH * scale_x;
		sizeHints.height = VIEW_HEIGHT * scale_y;
	}
	center_dx = (sizeHints.width - VIEW_WIDTH * scale_x) / 2;
	center_dy = (sizeHints.height - VIEW_HEIGHT * scale_y) / 2;

	sizeHints.flags = PSize | PMinSize | PMaxSize;
	sizeHints.min_width = sizeHints.width;
	sizeHints.min_height = sizeHints.height;
	sizeHints.max_width = sizeHints.width;
	sizeHints.max_height = sizeHints.height;
	if(!arg_fullscreen && arg_geometry) {
		int dummy;
		if(XWMGeometry(display, visualInfo.screen, arg_geometry, 0, 2, &sizeHints, &sizeHints.x, &sizeHints.y, &dummy, &dummy, &sizeHints.win_gravity) & (XValue | YValue)) sizeHints.flags |= USPosition | PWinGravity;
	}

	setWindowAttributes.background_pixel = blackPixel;
	setWindowAttributes.colormap = privateColourmap ? privateColourmap : defaultColourmap;
	window = XCreateWindow(display, RootWindow(display, visualInfo.screen), sizeHints.x, sizeHints.y, sizeHints.width, sizeHints.height, 2, visualInfo.depth, InputOutput, visualInfo.visual, CWBackPixel | CWColormap, &setWindowAttributes);
	if(window == 0) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, displayName, "failed to open window");
		return(1);
	}

	setWindowAttributes.backing_store = NotUseful;
	setWindowAttributes.save_under = False;
	setWindowAttributes.event_mask = KeyPressMask | KeyReleaseMask | EnterWindowMask | LeaveWindowMask | PointerMotionMask | KeymapStateMask | ExposureMask | FocusChangeMask;
	setWindowAttributes.override_redirect = arg_fullscreen ? True : False;
	XChangeWindowAttributes(display, window, CWBackingStore | CWSaveUnder | CWEventMask | CWOverrideRedirect, &setWindowAttributes);

	wmHints.flags = InputHint | StateHint;
	wmHints.input = True;
	wmHints.initial_state = NormalState;
	wmHints.icon_pixmap = XCreateBitmapFromData(display, window, icon_xbm_bits, icon_xbm_width, icon_xbm_height);
	wmHints.icon_mask = XCreateBitmapFromData(display, window, mask_xbm_bits, mask_xbm_width, mask_xbm_height);
	if(wmHints.icon_pixmap && wmHints.icon_mask) wmHints.flags |= IconPixmapHint | IconMaskHint;
	XmbSetWMProperties(display, window, "X11 Emerald Mine", global, 0, 0, &sizeHints, &wmHints, 0);

	deleteAtom = XInternAtom(display, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(display, window, &deleteAtom, 1);

	cursorBitmap = XCreateBitmapFromData(display, window, "", 1, 1);
	if(cursorBitmap) {
		allocColour.red = 0;
		allocColour.green = 0;
		allocColour.blue = 0;
		cursor = XCreatePixmapCursor(display, cursorBitmap, cursorBitmap, &allocColour, &allocColour, 0, 0);
	}

	XMapRaised(display, window);

	return(0);
}

/* handle frame sync, events */

int keyboard_raw = -1;
int keyboard_cook = -1;

static char keyboardVector[32];

static long frame_usec;

static int display_block;

static int cursor_timer;
static int cursor_hidden;

static XComposeStatus composeStatus;

void display_read(void)
{
	long usec;
	struct timeval tv;
	int kc;
	char buf[3];
	XEvent event;

	XSync(display, False); /* block until all graphics are drawn */

	if(display_block) XPeekEvent(display, &event); /* block until an event arrives */

	while(XPending(display)) {
		XNextEvent(display, &event);
		switch(event.xany.type) {
		case KeyPress: /* event.xkey; */
			kc = event.xkey.keycode & 255;
			keyboardVector[kc >> 3] |= (1 << (kc & 7));
			keyboard_raw = kc;
			buf[0] = 0; buf[1] = 0; buf[2] = 0;
			XLookupString(&event.xkey, buf, sizeof(buf), 0, &composeStatus);
			keyboard_cook = (buf[0] & 255) | (buf[1] & 255) << 8 | (buf[2] & 255) << 16;
			break;
		case KeyRelease: /* event.xkey; */
			kc = event.xkey.keycode & 255;
			keyboardVector[kc >> 3] &= ~(1 << (kc & 7));
			keyboard_raw = -1;
			keyboard_cook = -1;
			break;
		case MotionNotify: /* event.xmotion; */
			if(event.xmotion.x >= center_dx && event.xmotion.x <= center_dx + VIEW_WIDTH * scale_x && event.xmotion.y >= center_dy && event.xmotion.y <= center_dy + VIEW_HEIGHT * scale_y) {
				display_block &= ~4;
			} else {
				display_block |= 4;
			}
			cursor_timer = 32;
			break;
		case EnterNotify: /* event.xcrossing; */
			if((display_block & 1) != 0) {
				if(privateColourmap) XInstallColormap(display, privateColourmap);
			}
			display_block &= ~1;
			if(arg_fullscreen) XSetInputFocus(display, window, RevertToPointerRoot, CurrentTime);
			break;
		case LeaveNotify: /* event.xcrossing; */
			if((display_block & 1) == 0) {
				if(privateColourmap) XUninstallColormap(display, privateColourmap);
			}
			display_block |= 1;
			break;
		case FocusIn: /* event.xfocus; */
			display_block &= ~2;
			break;
		case FocusOut: /* event.xfocus; */
			display_block |= 2;
			break;
		case KeymapNotify: /* event.xkeymap; */
			memcpy(keyboardVector, event.xkeymap.key_vector, 32);
			break;
		case Expose: /* event.xexpose; */
			break;
		}
	}
	if(cursor_timer == 0 && cursor_hidden == 0) {
		if(cursor) XDefineCursor(display, window, cursor);
		cursor_hidden = 1;
	}
	if(cursor_timer && cursor_hidden) {
		if(cursor) XUndefineCursor(display, window);
		cursor_hidden = 0;
	}
	if(cursor_timer) cursor_timer--;

	gettimeofday(&tv, 0);
	usec = frame_usec + 1000000;
	frame_usec = (tv.tv_usec + tv.tv_sec * 1000000) * 60;
	usec -= frame_usec;
	if(usec >= 0 && usec <= 1500000) {
		frame_usec += usec;
#if 1
		usleep(usec / 60);
#else
		fd_set fds[3];
		FD_ZERO(fds + 0); FD_ZERO(fds + 1); FD_ZERO(fds + 2);
		tv.tv_usec = usec / 60; tv.tv_sec = 0;
		select(0, fds + 0, fds + 1, fds + 2, &tv); /* more portable than usleep? */
#endif
	}
}

int display_close(void)
{
	if(window) { XDestroyWindow(display, window); window = 0; }
	if(cursor) { XFreeCursor(display, cursor); cursor = 0; }
	if(cursorBitmap) { XFreePixmap(display, cursorBitmap); cursorBitmap = 0; }
	if(wmHints.icon_pixmap) { XFreePixmap(display, wmHints.icon_pixmap); wmHints.icon_pixmap = 0; }
	if(wmHints.icon_mask) { XFreePixmap(display, wmHints.icon_mask); wmHints.icon_mask = 0; }
	if(paletteNumUsed) {
		if(privateColourmap == 0) XFreeColors(display, defaultColourmap, palettePixels, paletteNumUsed, 0);
		paletteNumUsed = 0;
	}
	if(privateColours) { free(privateColours); privateColours = 0; }
	if(privateColourmap) { XFreeColormap(display, privateColourmap); privateColourmap = 0; }
	if(display) { XCloseDisplay(display); display = 0; }
	return(0);
}

static unsigned long privateAllocColour(unsigned short red, unsigned short green, unsigned short blue)
{
	int exact = -1, close = -1, any = -1;
	long exact_sum = 1000, close_sum = 1000000, any_sum = 1000000;
	int i;
	for(i = 0; i < privateNumColours; i++) {
		long sum;
		sum = labs(privateColours[i].red - red);
		sum += labs(privateColours[i].green - green);
		sum += labs(privateColours[i].blue - blue);
		if(sum < exact_sum) { exact_sum = sum; exact = i; }
		if(sum < close_sum && privateColours[i].flags) { close_sum = sum; close = i; }
		if(sum < any_sum) { any_sum = sum; any = i; }
	}
	if(exact != -1) {
		privateColours[exact].flags = 0;
		return(privateColours[exact].pixel);
	}
	if(close != -1) {
		privateColours[close].red = red;
		privateColours[close].green = green;
		privateColours[close].blue = blue;
		XStoreColor(display, privateColourmap, &privateColours[close]);
		privateColours[close].flags = 0;
		return(privateColours[close].pixel);
	}
	if(any != -1) {
		return(privateColours[any].pixel);
	}
	return(blackPixel);
}

/* ---------------------------------------------------------------------- */

/* handle keyboard, key names */

char keyboard_input[2][INPUT_MAX];

int keyboard_keycode[2][INPUT_MAX];

static const KeySym defaultKeySym[INPUT_MAX][2] = {
	[INPUT_north] = { XK_Up, XK_W },
	[INPUT_east] = { XK_Right, XK_D },
	[INPUT_south] = { XK_Down, XK_S },
	[INPUT_west] = { XK_Left, XK_A },
	[INPUT_shoot] = { XK_Control_R, XK_Control_L },
	[INPUT_abort] = { XK_Escape, XK_Escape },
};

int keyboard_open(void)
{
	int ply, button;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			keyboard_keycode[ply][button] = XKeysymToKeycode(display, defaultKeySym[button][ply]);
		}
	}
	return(0);
}

void keyboard_read(void)
{
	int ply, button;
	for(ply = 0; ply < 2; ply++) {
		for(button = 0; button < INPUT_MAX; button++) {
			int kc = keyboard_keycode[ply][button] & 255;
			keyboard_input[ply][button] = keyboardVector[kc >> 3] >> (kc & 7) & 1;
		}
	}
}

int keyboard_close(void)
{
	return(0);
}

char *keyboard_to_name(int kc)
{
	char *kn = XKeysymToString(XKeycodeToKeysym(display, kc, 0));
	if(kn == 0) kn = "";
	return(kn);
}

int keyboard_to_keycode(char *kn)
{
	KeySym ks = XStringToKeysym(kn);
	int kc = XKeysymToKeycode(display, ks);
	if(ks == NoSymbol || kc == 0) kc = -1;
	return(kc);
}

/* ---------------------------------------------------------------------- */

/* handle joysticks */

#if (defined __linux)
#include <linux/joystick.h>
#endif
#if (defined __NetBSD__ || defined __OpenBSD__)
#include <machine/joystick.h>
#endif
#if (defined __FreeBSD__)
#include <sys/joystick.h>
#endif

char joystick_input[2][INPUT_MAX];

static int calibrate_n[2], calibrate_e[2], calibrate_s[2], calibrate_w[2];

static int joy_fd[2] = { -1, -1 };
static int joy_index[2];
static time_t joy_delay;

#if (defined __linux)
static struct js_event joy_event[16];
#endif
#if (defined __NetBSD__ || defined __FreeBSD__ || defined __OpenBSD__)
static struct joystick joy_event[1];
#endif

int joystick_open(void)
{
	return(0);
}

void joystick_read(void)
{
#if (defined __linux || defined __NetBSD__ || defined __FreeBSD__ || defined __OpenBSD__)
	int ply;
	int len, i;
	for(ply = 0; ply < 2; ply++) {
		char *joy = joystick_input[ply];

/* attempt to open joystick devices. dont open the same device twice.
 * if none open, wait 2 seconds then try again. this should handle
 * joysticks plugged in while the game is running.
 */
		if(joy_fd[ply] == -1) {
			time_t joy_time = time(0);
			if((unsigned)(joy_time - joy_delay) > 2) {
				joy_delay = joy_time;
				for(i = 0; i < arg_joystick.length; i++) {
					if(joy_fd[1-ply] == -1 || joy_index[1-ply] != joy_index[ply]) { /* test for same device */
						joy_fd[ply] = open(arg_joystick.name[joy_index[ply]], O_RDONLY | O_NONBLOCK);
					}
					if(joy_fd[ply] != -1) break;
					joy_index[ply] = (joy_index[ply] + 1) % arg_joystick.length;
				}
			}
			calibrate_n[ply] = 32767;
			calibrate_e[ply] = -32767;
			calibrate_s[ply] = -32767;
			calibrate_w[ply] = 32767;
		}

		len = -1;
		if(joy_fd[ply] != -1) {
			len = read(joy_fd[ply], joy_event, sizeof(joy_event));
			if(len == -1) {
				if(errno != EAGAIN && errno != EINTR) {
					fprintf(stderr, "%s[%s] %s: %s\n", global, local, arg_joystick.name[joy_index[ply]], strerror(errno));
					close(joy_fd[ply]); joy_fd[ply] = -1;
				}
			}
		}
		len /= (int)(sizeof(*joy_event));

#if (defined __linux)
		for(i = 0; i < len; i++) {
			struct js_event *event = &joy_event[i];
			int val = event->value;
			switch(event->type & ~JS_EVENT_INIT) {
			case JS_EVENT_AXIS:
				if(event->number & 1) {
					if(calibrate_n[ply] > val) calibrate_n[ply] = val;
					if(calibrate_s[ply] < val) calibrate_s[ply] = val;
					joy[INPUT_north] = (val < (calibrate_n[ply] * 3 + calibrate_s[ply]) / 4);
					joy[INPUT_south] = (val > (calibrate_n[ply] + calibrate_s[ply] * 3) / 4);
				} else {
					if(calibrate_e[ply] < val) calibrate_e[ply] = val;
					if(calibrate_w[ply] > val) calibrate_w[ply] = val;
					joy[INPUT_east] = (val > (calibrate_e[ply] * 3 + calibrate_w[ply]) / 4);
					joy[INPUT_west] = (val < (calibrate_e[ply] + calibrate_w[ply] * 3) / 4);
				}
				break;
			case JS_EVENT_BUTTON:
				joy[INPUT_shoot] = (val != 0);
				break;
			}
		}
#endif
#if (defined __NetBSD__ || defined __FreeBSD__ || defined __OpenBSD__)
		for(i = 0; i < len; i++) {
			struct joystick *event = &joy_event[i];
			joy[INPUT_north] = (event->y < 64);
			joy[INPUT_east] = (event->x > 192);
			joy[INPUT_south] = (event->y > 192);
			joy[INPUT_west] = (event->x < 64);
			joy[INPUT_shoot] = ((event->b1 | event->b2) != 0);
		}
#endif
	}
#endif
}

int joystick_close(void)
{
	int ply;
	for(ply = 0; ply < 2; ply++) {
		if(joy_fd[ply] != -1) { close(joy_fd[ply]); joy_fd[ply] = -1; }
	}
	return(0);
}

/* ---------------------------------------------------------------------- */

/* handle audio */

/* advanced linux sound architecture */

#if (defined __linux || defined __QNX__)
#include <alsa/asoundlib.h>
#include <pthread.h>
#endif

/* open sound system */

#if (defined __linux)
#include <sys/soundcard.h>
#endif
#if (defined __NetBSD__ || defined __OpenBSD__ || defined __sun)
#include <sys/audioio.h>
#endif
#if (defined __FreeBSD__)
#include <soundcard.h>
#endif
#if (defined __hpux)
#include <sys/audio.h>
#endif

static char sound_play[SOUND_MAX]; /* should probably be protected by a semaphore */
static int sound_ptr[SOUND_MAX]; /* play position in the sound */

#ifdef SND_LIB_VERSION /* alsa */

static void *audio_thread(void *arg);
static pthread_t audio_pid;
static pthread_cond_t audio_pipe_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t audio_pipe_mutex = PTHREAD_MUTEX_INITIALIZER;

int audio_open(void)
{
	if(errno = pthread_create(&audio_pid, 0, audio_thread, 0), errno) {
		audio_pid = 0;
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to create thread", strerror(errno));
		return(1);
	}
	return(0);
}

void audio_write(struct LOGIC *lev)
{
	int i;
	for(i = 0; i < SOUND_MAX; i++) if(lev->sound[i]) {
		sound_play[i] = 1;
	}
	pthread_cond_signal(&audio_pipe_cond);
}

int audio_close(void)
{
	if(audio_pid) {
		pthread_cancel(audio_pid);
		pthread_join(audio_pid, 0);
		audio_pid = 0;
	}
	return(0);
}

#else /* oss */

static int audio_thread(void);
static pid_t audio_pid = -1;
static int audio_pipe[2] = { -1, -1 };

int audio_open(void)
{
	signal(SIGPIPE, SIG_IGN); /* dont crash if sound process dies */

	if(pipe(audio_pipe) == -1) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to create pipe", strerror(errno));
		return(1);
	}
	if(fcntl(audio_pipe[0], F_SETFL, O_NONBLOCK) == -1 || fcntl(audio_pipe[1], F_SETFL, O_NONBLOCK) == -1) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to nonblock pipe", strerror(errno));
		return(1);
	}
	audio_pid = fork();
	if(audio_pid == -1) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to fork process", strerror(errno));
		return(1);
	}
	if(audio_pid == 0) {
		audio_pid = -1;
		close(audio_pipe[1]); audio_pipe[1] = -1;
		_exit(audio_thread());
	}
	close(audio_pipe[0]); audio_pipe[0] = -1;

	return(0);
}

void audio_write(struct LOGIC *lev)
{
	int i;
	for(i = 0; i < SOUND_MAX; i++) if(lev->sound[i]) {
		sound_play[i] = 1;
	}
	if(audio_pipe[1] != -1) {
		if(write(audio_pipe[1], sound_play, SOUND_MAX) == -1) {
			if(errno != EAGAIN && errno != EINTR) {
				fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to write pipe", strerror(errno));
				close(audio_pipe[1]); audio_pipe[1] = -1;
			}
		}
	}
	memset(sound_play, 0, SOUND_MAX);
}

int audio_close(void)
{
	if(audio_pid != -1) {
		kill(audio_pid, SIGTERM);
		waitpid(audio_pid, 0, 0);
		audio_pid = -1;
	}
	if(audio_pipe[0] != -1) { close(audio_pipe[0]); audio_pipe[0] = -1; }
	if(audio_pipe[1] != -1) { close(audio_pipe[1]); audio_pipe[1] = -1; }

	return(0);
}

#endif

/* mix audio data */

static int mix_play[MIXER_MAX]; /* sounds we have chosen to play */
static int mix_count;

#define BUFFER_SIZE 256

static short buf_mix[BUFFER_SIZE]; /* buffer to mix sounds */
static char buf_out[BUFFER_SIZE]; /* buffer written to device */

static void audio_mix(void)
{
	int i;
	for(i = 0; i < mix_count; i++) {
		register const unsigned char *in = sounds_data[mix_play[i]] + sound_ptr[mix_play[i]];
		register short *out = buf_mix;
		register int len = sounds_length[mix_play[i]] - sound_ptr[mix_play[i]];
		if(len > BUFFER_SIZE) len = BUFFER_SIZE;
		do *out++ += ulaw_to_linear[*in++] / MIXER_MAX; while(--len);
	}
	{
		register const short *in = buf_mix;
		register char *out = buf_out;
		register int len = BUFFER_SIZE;
		do *out++ = linear_to_ulaw[(*in++ >> 2) & 16383]; while(--len);
	}
}

/* audio thread */

static int audio_index;
static time_t audio_delay;

#ifdef SND_LIB_VERSION /* alsa */

static snd_pcm_t *audio_fd;
static snd_pcm_hw_params_t *audioInfo;

static void shutup(void) { }

static void *audio_thread(void *arg)
{
	int timer = 0;
	int i;

	snd_lib_error_set_handler((snd_lib_error_handler_t)(shutup));
	snd_pcm_hw_params_alloca(&audioInfo);

	pthread_mutex_lock(&audio_pipe_mutex);

	memset(sound_ptr, -1, sizeof(sound_ptr));

	for(;;) {

/* start playing sounds */

		mix_count = 0;

		for(i = 0; i < SOUND_MAX; i++) if(sound_play[i]) {
			sound_play[i] = 0;
			sound_ptr[i] = 0;
		}
		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			while(mix_count < MIXER_MAX) {
				mix_play[mix_count++] = i;
				if(i >= SOUND_boom) break;
			}
		}
		if(mix_count) timer = 64; /* keep device open ~2 seconds */

/* read message from pipe */

		if(timer == 0) {
			pthread_cond_wait(&audio_pipe_cond, &audio_pipe_mutex);
			continue;
		}

		if(timer) timer--;

/* open audio device */

		if(timer && audio_fd == 0) {
			time_t audio_time = time(0);
			if((unsigned)(audio_time - audio_delay) > 2) {
				audio_delay = audio_time;
				for(i = 0; i < arg_audio.length; i++) {
					snd_pcm_open(&audio_fd, arg_audio.name[audio_index], SND_PCM_STREAM_PLAYBACK, 0);
					if(audio_fd) break;
					audio_index = (audio_index + 1) % arg_audio.length;
				}
			}
			if(audio_fd && audioInfo) {
				snd_pcm_uframes_t period_size = BUFFER_SIZE;
				unsigned int periods = 2;
				snd_pcm_hw_params_any(audio_fd, audioInfo);
				snd_pcm_hw_params_set_access(audio_fd, audioInfo, SND_PCM_ACCESS_RW_INTERLEAVED);
				snd_pcm_hw_params_set_format(audio_fd, audioInfo, SND_PCM_FORMAT_MU_LAW);
				snd_pcm_hw_params_set_channels(audio_fd, audioInfo, 1);
				snd_pcm_hw_params_set_rate(audio_fd, audioInfo, 8000, 0);
				snd_pcm_hw_params_set_period_size_near(audio_fd, audioInfo, &period_size, 0);
				snd_pcm_hw_params_set_periods_near(audio_fd, audioInfo, &periods, 0);
				snd_pcm_hw_params(audio_fd, audioInfo);
			}
		}
		if(timer == 0 && audio_fd) {
			snd_pcm_close(audio_fd); audio_fd = 0;
			audio_delay = 0;
		}

/* write audio data */

		memset(buf_mix, 0, sizeof(buf_mix));
		audio_mix();

		if(audio_fd) {
			if(i = snd_pcm_writei(audio_fd, buf_out, BUFFER_SIZE), i < 0) {
				if(i = snd_pcm_prepare(audio_fd), i < 0) {
					fprintf(stderr, "%s[%s] %s: %s\n", global, local, snd_pcm_name(audio_fd), snd_strerror(i));
					snd_pcm_close(audio_fd); audio_fd = 0;
				}
			}
		}

		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			sound_ptr[i] += BUFFER_SIZE;
			if(sound_ptr[i] >= sounds_length[i]) sound_ptr[i] = -1;
		}
	}
	return(arg);
}

#else /* oss */

static int audio_fd = -1;

static int audio_thread(void)
{
	int timer = 0;
	int i;

	char blocking = 0;

	memset(sound_ptr, -1, sizeof(sound_ptr));

	for(;;) {

/* start playing sounds */

		mix_count = 0;

		for(i = 0; i < SOUND_MAX; i++) if(sound_play[i]) {
			sound_play[i] = 0;
			sound_ptr[i] = 0;
		}
		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			while(mix_count < MIXER_MAX) {
				mix_play[mix_count++] = i;
				if(i >= SOUND_boom) break;
			}
		}
		if(mix_count) timer = 64; /* keep device open ~2 seconds */

/* read message from pipe */

		if(timer && blocking) {
			if(fcntl(audio_pipe[0], F_SETFL, O_NONBLOCK) == -1) {
				fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to nonblock pipe", strerror(errno));
				return(1);
			}
			blocking = 0;
		}
		if(timer == 0 && blocking == 0) {
			if(fcntl(audio_pipe[0], F_SETFL, 0) == -1) {
				fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to block pipe", strerror(errno));
				return(1);
			}
			blocking = 1;
		}
		i = read(audio_pipe[0], sound_play, SOUND_MAX);
		if(i == -1) {
			if(errno != EAGAIN && errno != EINTR) {
				fprintf(stderr, "%s[%s] %s: %s\n", global, local, "failed to read pipe", strerror(errno));
				return(1);
			}
		}
		if(i == SOUND_MAX) continue;

		if(timer) timer--;

/* open audio device */

		if(timer && audio_fd == -1) {
			time_t audio_time = time(0);
			if((unsigned)(audio_time - audio_delay) > 2) {
				audio_delay = audio_time;
				for(i = 0; i < arg_audio.length; i++) {
					audio_fd = open(arg_audio.name[audio_index], O_WRONLY);
					if(audio_fd != -1) break;
					audio_index = (audio_index + 1) % arg_audio.length;
				}
			}
			if(audio_fd != -1) {

/* setting the buffer size is necessary for a nice game,
 * so we need to hunt around for the ioctl. it will work
 * without it, but the sound wont be in sync.
 */

#ifdef SNDCTL_DSP_SETFRAGMENT
				int audioInfo;
				audioInfo = 0x00020008;
				ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &audioInfo);
				audioInfo = AFMT_MU_LAW;
				ioctl(audio_fd, SNDCTL_DSP_SETFMT, &audioInfo);
#endif
#ifdef AUDIO_INITINFO
				audio_info_t audioInfo;
				AUDIO_INITINFO(&audioInfo);
				audioInfo.play.buffer_size = 256;
				ioctl(audio_fd, AUDIO_SETINFO, &audioInfo);
#endif
#ifdef AUDIO_SET_TXBUFSIZE
				int audioInfo;
				audioInfo = 256;
				ioctl(audio_fd, AUDIO_SET_TXBUFSIZE, audioInfo);
#endif
			}
		}
		if(timer == 0 && audio_fd != -1) {
			close(audio_fd); audio_fd = -1;
			audio_delay = 0;
		}

/* write audio data */

		memset(buf_mix, 0, sizeof(buf_mix));
		audio_mix();

		if(audio_fd != -1) {
			if(write(audio_fd, buf_out, BUFFER_SIZE) == -1) {
				if(errno != EAGAIN && errno != EINTR) {
					fprintf(stderr, "%s[%s] %s: %s\n", global, local, arg_audio.name[audio_index], strerror(errno));
					close(audio_fd); audio_fd = -1;
				}
			}
		}

		for(i = 0; i < SOUND_MAX; i++) if(sound_ptr[i] != -1) {
			sound_ptr[i] += BUFFER_SIZE;
			if(sound_ptr[i] >= sounds_length[i]) sound_ptr[i] = -1;
		}
	}
	return(0);
}

#endif

/* ---------------------------------------------------------------------- */

/* create pixmaps, create graphics contexts */

static Pixmap screenPixmap;
static Pixmap objectPixmap;
static Pixmap objectBitmap;
static Pixmap scorePixmap;
static Pixmap scoreBitmap;
static Pixmap spriteBitmap;

static GC pixmapGC;
static GC bitmapGC;

static XGCValues gcValues;

int screen_open(void)
{
	screenPixmap = XCreatePixmap(display, window, (VIEW_WIDTH + 16) * scale_x, (VIEW_HEIGHT + 16) * scale_y, visualInfo.depth);
	if(screenPixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create screen pixmap", (VIEW_WIDTH + 16) * scale_x, (VIEW_HEIGHT + 16) * scale_y, visualInfo.depth);
		return(1);
	}
	objectPixmap = XCreatePixmap(display, window, GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, visualInfo.depth);
	if(objectPixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create graphics pixmap", GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, visualInfo.depth);
		return(1);
	}
	objectBitmap = XCreatePixmap(display, window, GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, 1);
	if(objectBitmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create graphics bitmap", GFX_WIDTH * scale_x, GFX_HEIGHT * scale_y, 1);
		return(1);
	}
	scorePixmap = XCreatePixmap(display, window, SCORE_WIDTH * scale_x, 8 * scale_y, visualInfo.depth);
	if(scorePixmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create score pixmap", SCORE_WIDTH * scale_x, 8 * scale_y, visualInfo.depth);
		return(1);
	}
	scoreBitmap = XCreatePixmap(display, window, SCORE_WIDTH * scale_x, 8 * scale_y, 1);
	if(scoreBitmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create score bitmap", SCORE_WIDTH * scale_x, 8 * scale_y, 1);
		return(1);
	}
	spriteBitmap = XCreatePixmap(display, window, 16 * scale_x, 16 * scale_y, 1);
	if(spriteBitmap == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create sprite bitmap", 16 * scale_x, 16 * scale_y, 1);
		return(1);
	}

	gcValues.function = GXcopy;
	gcValues.foreground = 1;
	gcValues.background = 0;
	gcValues.line_width = 1;
	gcValues.line_style = LineDoubleDash;
	gcValues.cap_style = CapProjecting;
	gcValues.fill_style = FillSolid;
	gcValues.graphics_exposures = False;
	pixmapGC = XCreateGC(display, objectPixmap, GCFunction | GCLineWidth | GCLineStyle | GCCapStyle | GCFillStyle | GCGraphicsExposures, &gcValues);
	if(pixmapGC == 0) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, displayName, "failed to create pixmap graphics context");
		return(1);
	}
	bitmapGC = XCreateGC(display, objectBitmap, GCFunction | GCForeground | GCBackground | GCGraphicsExposures, &gcValues);
	if(bitmapGC == 0) {
		fprintf(stderr, "%s[%s] %s: %s\n", global, local, displayName, "failed to create bitmap graphics context");
		return(1);
	}

	return(screen_write());
}

/* render graphics */

static XImage *pixmapImage;
static XImage *bitmapImage;

int screen_write(void)
{
	int top, bottom, inc = 64; /* render in parts to save memory */
	int x, in_x, out_x;
	int y, in_y, out_y;

	int pixmap_bytes, bitmap_bytes;
	char *pixmap_data, *bitmap_data;
	char *pixmap_black;

	pixmapImage = XCreateImage(display, visualInfo.visual, visualInfo.depth, ZPixmap, 0, 0, GFX_WIDTH * scale_x, inc * scale_y + 1, BitmapPad(display), 0);
	if(pixmapImage == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create pixmap image", GFX_WIDTH * scale_x, inc * scale_y + 1, visualInfo.depth);
		return(1);
	}
	pixmapImage->data = malloc(pixmapImage->height * pixmapImage->bytes_per_line);
	if(pixmapImage->data == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%d)\n", global, local, displayName, "failed to allocate memory for pixmap image", pixmapImage->height * pixmapImage->bytes_per_line);
		return(1);
	}
	bitmapImage = XCreateImage(display, visualInfo.visual, 1, XYBitmap, 0, 0, GFX_WIDTH * scale_x, inc * scale_y, BitmapPad(display), 0);
	if(bitmapImage == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%dx%dx%d)\n", global, local, displayName, "failed to create bitmap image", GFX_WIDTH * scale_x, inc * scale_y, 1);
		return(1);
	}
	bitmapImage->data = malloc(bitmapImage->height * bitmapImage->bytes_per_line);
	if(bitmapImage->data == 0) {
		fprintf(stderr, "%s[%s] %s: %s (%d)\n", global, local, displayName, "failed to allocate memory for bitmap image", bitmapImage->height * bitmapImage->bytes_per_line);
		return(1);
	}

	pixmap_bytes = pixmapImage->bytes_per_line;
	bitmap_bytes = bitmapImage->bytes_per_line;
	pixmap_data = pixmapImage->data;
	bitmap_data = bitmapImage->data;
	pixmap_black = pixmap_data + inc * scale_y * pixmap_bytes;

	for(x = 0; x < GFX_WIDTH * scale_x; x++) XPutPixel(pixmapImage, x, inc * scale_y, blackPixel); /* cache a line of black */

	for(top = 0; top < GFX_HEIGHT; top = bottom) {
		bottom = top + inc; if(bottom > GFX_HEIGHT) bottom = GFX_HEIGHT;
		out_y = 0;
		for(in_y = top; in_y < bottom; in_y++) {
			char *pixmap_ptr = pixmap_data + out_y * pixmap_bytes;
			char *bitmap_ptr = bitmap_data + out_y * bitmap_bytes;
			memcpy(pixmap_ptr, pixmap_black, pixmap_bytes);
			memset(bitmap_ptr, 0, bitmap_bytes);
			out_x = 0;
			for(in_x = 0; in_x < GFX_WIDTH; in_x++) {
				char c = graphics_xpm[in_x][in_y];
				unsigned long pixel = paletteHash[c & 127];
				for(x = 0; x < scale_x; x++) {
					if(pixel != blackPixel) XPutPixel(pixmapImage, out_x, out_y, pixel);
					if(c != ' ') XPutPixel(bitmapImage, out_x, out_y, 1);
					out_x++;
				}
			}
			for(y = 1; y < scale_y; y++) {
				memcpy(pixmap_ptr + y * pixmap_bytes, pixmap_ptr, pixmap_bytes);
				memcpy(bitmap_ptr + y * bitmap_bytes, bitmap_ptr, bitmap_bytes);
			}
			out_y += scale_y;
		}
		XPutImage(display, objectPixmap, pixmapGC, pixmapImage, 0, 0, 0, top * scale_y, GFX_WIDTH * scale_x, (bottom - top) * scale_y);
		XPutImage(display, objectBitmap, bitmapGC, bitmapImage, 0, 0, 0, top * scale_y, GFX_WIDTH * scale_x, (bottom - top) * scale_y);
	}
	XDestroyImage(pixmapImage); pixmapImage = 0;
	XDestroyImage(bitmapImage); bitmapImage = 0;

	return(0);
}

int screen_close(void)
{
	if(pixmapImage) { XDestroyImage(pixmapImage); pixmapImage = 0; }
	if(bitmapImage) { XDestroyImage(bitmapImage); bitmapImage = 0; }
	if(pixmapGC) { XFreeGC(display, pixmapGC); pixmapGC = 0; }
	if(bitmapGC) { XFreeGC(display, bitmapGC); bitmapGC = 0; }
	if(screenPixmap) { XFreePixmap(display, screenPixmap); screenPixmap = 0; }
	if(objectPixmap) { XFreePixmap(display, objectPixmap); objectPixmap = 0; }
	if(objectBitmap) { XFreePixmap(display, objectBitmap); objectBitmap = 0; }
	if(scorePixmap) { XFreePixmap(display, scorePixmap); scorePixmap = 0; }
	if(scoreBitmap) { XFreePixmap(display, scoreBitmap); scoreBitmap = 0; }
	if(spriteBitmap) { XFreePixmap(display, spriteBitmap); spriteBitmap = 0; }
	return(0);
}

/* ---------------------------------------------------------------------- */

/* game drawing functions */

static short screen_cache[VIEW_WIDTH / 16 + 1][VIEW_HEIGHT / 16 + 1];

static int scn_left;
static int scn_top;

static int scn_pixel_left;
static int scn_pixel_top;

void x11_game_clear(void)
{
	XSetForeground(display, pixmapGC, blackPixel);
	XFillRectangle(display, screenPixmap, pixmapGC, 0, 0, (VIEW_WIDTH + 16) * scale_x, (VIEW_HEIGHT + 16) * scale_y);
	scn_left = 0;
	scn_top = 0;
	memset(screen_cache, -1, sizeof(screen_cache)); /* mark entire screen as dirty */
}

void x11_game_scroll(struct LOGIC *lev)
{
	int width = 0, src_left = 0, dst_left = 0;
	int height = 0, src_top = 0, dst_top = 0;

	int direction = 0;

	int x, src_x, dst_x;
	int y, src_y, dst_y;

	scn_pixel_left = lev->pixel_left % 16;
	scn_pixel_top = lev->pixel_top % 16;

	if(lev->view_top > scn_top) {
		height = lev->view_top - scn_top;
		src_top = height;
		direction -= 1; /* north */
	}
	if(lev->view_top < scn_top) {
		height = scn_top - lev->view_top;
		dst_top = height;
		direction += 1; /* south */
	}
	if(lev->view_left > scn_left) {
		width = lev->view_left - scn_left;
		src_left = width;
		direction -= 3; /* west */
	}
	if(lev->view_left < scn_left) {
		width = scn_left - lev->view_left;
		dst_left = width;
		direction += 3; /* east */
	}
	scn_left = lev->view_left;
	scn_top = lev->view_top;

	if(direction == 0 || width >= VIEW_WIDTH / 16 + 1 || height >= VIEW_HEIGHT / 16 + 1) return; /* nothing to do */
	width = (VIEW_WIDTH / 16 + 1) - width;
	height = (VIEW_HEIGHT / 16 + 1) - height;

	if(direction < 0) {
		src_x = src_left; dst_x = dst_left;
		for(x = 0; x < width; x++) {
			src_y = src_top; dst_y = dst_top;
			for(y = 0; y < height; y++) {
				screen_cache[dst_x][dst_y] = screen_cache[src_x][src_y];
				src_y++; dst_y++;
			}
			src_x++; dst_x++;
		}
	} else {
		src_x = src_left + width; dst_x = dst_left + width;
		for(x = 0; x < width; x++) {
			src_x--; dst_x--;
			src_y = src_top + height; dst_y = dst_top + height;
			for(y = 0; y < height; y++) {
				src_y--; dst_y--;
				screen_cache[dst_x][dst_y] = screen_cache[src_x][src_y];
			}
		}
	}

	XCopyArea(display, screenPixmap, screenPixmap, pixmapGC, src_left * 16 * scale_x, src_top * 16 * scale_y, width * 16 * scale_x, height * 16 * scale_y, dst_left * 16 * scale_x, dst_top * 16 * scale_y);
	XFlush(display);
}

void x11_game_object(struct LOGIC *lev)
{
	int tile, tile_x, tile_y;
	int scn_x, lev_x;
	int scn_y, lev_y;
	scn_x = 0;
	for(lev_x = lev->view_left; lev_x <= lev->view_right; lev_x++) {
		scn_y = 0;
		for(lev_y = lev->view_top; lev_y <= lev->view_bottom; lev_y++) {
			tile = object_map[lev->cave[lev_x][lev_y]][lev->frame];
			if(screen_cache[scn_x][scn_y] != tile) {
				screen_cache[scn_x][scn_y] = tile;
				tile_x = (tile / GFX_HEIGHT) * 16;
				tile_y = (tile % GFX_HEIGHT);
				XCopyArea(display, objectPixmap, screenPixmap, pixmapGC, tile_x * scale_x, tile_y * scale_y, 16 * scale_x, 16 * scale_y, scn_x * 16 * scale_x, scn_y * 16 * scale_y);
			}
			scn_y++;
		}
		scn_x++;
	}
}

void x11_game_player(struct LOGIC *lev)
{
	int tile, tile_x, tile_y;
	int dx, x0, x1;
	int dy, y0, y1;
	int ply;
	for(ply = 0; ply < 2; ply++) {
		if(lev->hide[ply]) continue;

		x0 = lev->pl_pixel_left[ply];
		x1 = lev->pl_pixel_right[ply];
		y0 = lev->pl_pixel_top[ply];
		y1 = lev->pl_pixel_bottom[ply];
		if(!(x0 >= lev->pixel_left && x1 <= lev->pixel_right && y0 >= lev->pixel_top && y1 <= lev->pixel_bottom)) continue;

		XSetFunction(display, bitmapGC, GXcopyInverted);
		dx = x0 % 16; dy = y0 % 16;
		tile = object_map[lev->cave[x0 / 16][y0 / 16]][lev->frame];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		XCopyArea(display, objectBitmap, spriteBitmap, bitmapGC, (tile_x + dx) * scale_x, (tile_y + dy) * scale_y, (16 - dx) * scale_x, (16 - dy) * scale_y, 0, 0);
		dx = 15 - x1 % 16; dy = 15 - y1 % 16;
		tile = object_map[lev->cave[x1 / 16][y1 / 16]][lev->frame];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		XCopyArea(display, objectBitmap, spriteBitmap, bitmapGC, tile_x * scale_x, tile_y * scale_y, (16 - dx) * scale_x, (16 - dy) * scale_y, dx * scale_x, dy * scale_y);
		XSetFunction(display, bitmapGC, GXcopy);

		x0 -= lev->pixel_left - scn_pixel_left;
		x1 -= lev->pixel_left - scn_pixel_left;
		y0 -= lev->pixel_top - scn_pixel_top;
		y1 -= lev->pixel_top - scn_pixel_top;

		screen_cache[x0 / 16][y0 / 16] = -1; /* mark screen over player as dirty */
		screen_cache[x1 / 16][y1 / 16] = -1;

		tile = sprite_map[lev->pl_anim[ply]][ply][lev->frame];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		XSetClipMask(display, pixmapGC, spriteBitmap);
		XSetClipOrigin(display, pixmapGC, x0 * scale_x, y0 * scale_y);
		XCopyArea(display, objectPixmap, screenPixmap, pixmapGC, tile_x * scale_x, tile_y * scale_y, 16 * scale_x, 16 * scale_y, x0 * scale_x, y0 * scale_y);
		XSetClipMask(display, pixmapGC, None);
	}
}

void x11_game_score(struct LOGIC *lev)
{
	int x = scn_pixel_left, x0, x1;
	int y = scn_pixel_top + 4, y0, y1;

	XSetClipMask(display, pixmapGC, scoreBitmap);
	XSetClipOrigin(display, pixmapGC, x * scale_x, y * scale_y);
	XCopyArea(display, scorePixmap, screenPixmap, pixmapGC, 0, 0, SCORE_WIDTH * scale_x, 8 * scale_y, x * scale_x, y * scale_y);
	XSetClipMask(display, pixmapGC, None);

	x0 = x / 16; x1 = (x + SCORE_WIDTH - 1) / 16;
	y0 = y / 16; y1 = (y + 8 - 1) / 16;
	for(x = x0; x <= x1; x++) for(y = y0; y <= y1; y++) {
		screen_cache[x][y] = -1; /* mark screen under score as dirty */
	}
}

void x11_game_to_screen(void)
{
	XCopyArea(display, screenPixmap, window, pixmapGC, scn_pixel_left * scale_x, scn_pixel_top * scale_y, VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y, center_dx, center_dy);
	XFlush(display);
}

/* ---------------------------------------------------------------------- */

/* score drawing functions */

static short score_cache[SCORE_WIDTH / 8];

void x11_score_clear(void)
{
	XSetForeground(display, pixmapGC, blackPixel);
	XFillRectangle(display, scorePixmap, pixmapGC, 0, 0, SCORE_WIDTH * scale_x, 8 * scale_y);
	XFillRectangle(display, scoreBitmap, bitmapGC, 0, 0, SCORE_WIDTH * scale_x, 8 * scale_y);
	memset(score_cache, -1, sizeof(score_cache));
}

void x11_score_print(char str[40])
{
	int tile, tile_x, tile_y;
	int x;
	for(x = 0; x < SCORE_WIDTH / 8; x++) {
		tile = score_map[str[x] & 127];
		if(score_cache[x] != tile) {
			score_cache[x] = tile;
			tile_x = (tile / GFX_HEIGHT) * 16;
			tile_y = (tile % GFX_HEIGHT);
			XCopyArea(display, objectPixmap, scorePixmap, pixmapGC, tile_x * scale_x, tile_y * scale_y, 8 * scale_x, 8 * scale_y, x * 8 * scale_x, 0);
			XCopyArea(display, objectBitmap, scoreBitmap, bitmapGC, tile_x * scale_x, tile_y * scale_y, 8 * scale_x, 8 * scale_y, x * 8 * scale_x, 0);
		}
	}
}

/* ---------------------------------------------------------------------- */

/* menu drawing functions */

static int antsPos;
static char antsDashes[2];

void x11_menu_clear(void)
{
	XSetForeground(display, pixmapGC, palettePixels[60]);
	XFillRectangle(display, screenPixmap, pixmapGC, 0, 0, VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y);
}

void x11_menu_print(char str[26], int top)
{
	int tile, tile_x, tile_y;
	int x;
	XSetForeground(display, pixmapGC, palettePixels[60]);
	XFillRectangle(display, screenPixmap, pixmapGC, 0, top * 16 * scale_y, VIEW_WIDTH * scale_x, 16 * scale_y);
	for(x = 0; x < 26; x++) {
		tile = font_map[str[x] & 127];
		tile_x = (tile / GFX_HEIGHT) * 16;
		tile_y = (tile % GFX_HEIGHT);
		XCopyArea(display, objectBitmap, spriteBitmap, bitmapGC, tile_x * scale_x, tile_y * scale_y, 16 * scale_x, 16 * scale_y, 0, 0);
		XSetClipMask(display, pixmapGC, spriteBitmap);
		XSetClipOrigin(display, pixmapGC, (x * 12 + 4) * scale_x, top * 16 * scale_y);
		XCopyArea(display, objectPixmap, screenPixmap, pixmapGC, tile_x * scale_x, tile_y * scale_y, 16 * scale_x, 16 * scale_y, (x * 12 + 4) * scale_x, top * 16 * scale_y);
		XSetClipMask(display, pixmapGC, None);
	}
}

void x11_menu_ants(int left, int right, int top)
{
	int x, x0 = left * scale_x, x1 = (right + 1) * scale_x - 1;
	int y, y0 = top * 16 * scale_y, y1 = (top + 1) * 16 * scale_y - 1;

	antsPos = (antsPos + 1) % 16;

	XSetForeground(display, pixmapGC, palettePixels[COLOUR_white]);
	XSetBackground(display, pixmapGC, palettePixels[COLOUR_red]);
	antsDashes[0] = 8 * scale_y; antsDashes[1] = 8 * scale_y;
	XSetDashes(display, pixmapGC, antsPos * scale_y, antsDashes, 2);
	for(x = 0; x < scale_x; x++) {
		XDrawLine(display, screenPixmap, pixmapGC, x1 - x, y0, x1 - x, y1);
		XDrawLine(display, screenPixmap, pixmapGC, x0 + x, y1, x0 + x, y0);
	}
	antsDashes[0] = 8 * scale_x; antsDashes[1] = 8 * scale_x;
	XSetDashes(display, pixmapGC, antsPos * scale_x, antsDashes, 2);
	for(y = 0; y < scale_y; y++) {
		XDrawLine(display, screenPixmap, pixmapGC, x0, y0 + y, x1, y0 + y);
		XDrawLine(display, screenPixmap, pixmapGC, x1, y1 - y, x0, y1 - y);
	}
}

void x11_menu_to_screen(void)
{
	XCopyArea(display, screenPixmap, window, pixmapGC, 0, 0, VIEW_WIDTH * scale_x, VIEW_HEIGHT * scale_y, center_dx, center_dy);
	XFlush(display);
}
